
#include <Arduino.h>
#include <soc/syscon_reg.h>
#include <OneWire.h>
#include <Wire.h>
#include <U8g2lib.h>
#include <Adafruit_ADS1015.h>                         // Библиотека для работы с модулями ADS1115 и ADS1015  
#include <PID_v1.h>
#include <ESP32Encoder.h>
#include <Preferences.h>

const byte Versiya[]  = {30, 3, 24};

Preferences Prefs;

float  Temp[6];
byte Write  = 0;

Adafruit_ADS1115 ads; //  0x48                        // Создание объекта ads для общения через него с модулем 

int16_t Adc[] = {0, 0, 0, 0};                         // Переменная для хранения значения АЦП вольтметра (канал 0-3)
float   Volt_Value[] = {0, 0, 0, 0}; 

#define SDA 21
#define SCL 22
U8G2_ST7567_ENH_DG128064I_1_HW_I2C u8g2(U8G2_R0, SCL, SDA, U8X8_PIN_NONE); // ESP32

hw_timer_t *Timer0_Cfg = NULL;
hw_timer_t *Timer2_Cfg = NULL;

double Setpoint[2], Input[2], Output[2];
double Kp[] = {1220.0,  3.0};
double Ki[] = {2250.0, 20.0};
double Kd[] = {   5.5,  0.5};
PID PID_CUR(&Input[0], &Output[0], &Setpoint[0], Kp[0], Ki[0], Kd[0], DIRECT);
PID PID_ZAR(&Input[1], &Output[1], &Setpoint[1], Kp[1], Ki[1], Kd[1], DIRECT);
/*-----------------------------------------------------------------------------------------------------*/
ESP32Encoder encoder;

#define CLK               13                // CLK ENCODER 
#define DT                14                // DT  ENCODER 
#define SB_Enc            27                // SW  ENCODER
long  oldPosition       = 0;
/*-----------------------------------------------------------------------------------------------------*/
const int Var_L         = 4094;
float Volt[]            = {0, 0};
//float Volt_Contr[]      = {12.0, 0};
float Volt_Lim[]        = {0, 0, 0, 8.45, 12.7, 16.95, 21.2, 25.35, 16.5, 14.0};
//                        {0, 0, 0, 8.4,  12.6, 16.8,  21.0, 25.2,  12.6, 12.7};
int   Volt_Min;
int   Tmp_V[]           = {Var_L, Var_L, 2004, 3784, 3444, 3114, 2834, 2534, 3114, 3114};
float Curr[]            = {0.0, 0.0, 1.0, 1.5, 4.0, 3.0};
float Curr_T            = 0.0;
float Curr_End          = 0.0;
byte  Select            = 0;
int   Level             = Var_L;
byte  Zar               = 0;
byte  Second[]          = {0};
byte  Minute[]          = {0};
byte  Hour[]            = {0}; 
int   Sec[]             = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
byte  Dw                = 0;
/*-----------------------------------------------------------------------------------------------------*/
#define But_Zar           35
#define Beep              32
byte  Beep_ON           = 0;
byte  Beep_OFF          = 0;

const int Pwm_Freq      = 5000;
const int Pwm_Bit       = 12;
const int Pwm_Channel   = 0;
const int Fan_Channel   = 1;

#define PWM_OUT           17
#define Fan_PWM           16

int   Duration          = 1;
byte  Flag_Plum[]       = {0, 0};
int   Ind_C             = 30;
int   Sp_Pwm            = 0;
byte  Sp[3];
byte  Del_V             = 0;
/*-----------------------------------------------------------------------------------------------------*/
OneWire DS_BP(4);                                     // Pin подключения датчика температуры радиатора

#define POWER_MODE        0                           // Режим питания, 0 - внешнее, 1 - паразитное
boolean FlagSensReady_BP;                             // Признак готовности данных с датчика
byte  bufData_BP[9];                                  // Буфер данных
int   TimeCount;                                      // Счетчик времени измерения
byte  Present_BP        = 0;                          // Сигнал наличия термодатчика
byte  Flag_Cels         = 0;                          // Флаг готовности данных

byte  Cels_BP[]         = {0};
byte  Cls[2];
/*-----------------------------------------------------------------------------------------------------*/
#define Zero_Time  Sec[0] = 0; Second[0] = 0; Minute[0] = 0; Hour[0] = 0;
#define LOAD_W    Temp[0] = (float)Level; Temp[1] = Curr[2]; Temp[2] = Curr[3]; Temp[3] = Curr[4]; Temp[4] = Curr[5];
#define LOAD_R      Level = (int)Temp[0]; Curr[2] = Temp[1]; Curr[3] = Temp[2]; Curr[4] = Temp[3]; Curr[5] = Temp[4];

void Volt_M(byte Num)   {Volt_Min = (Var_L - (3 * Num) * 90); Zar = 0; Level = Tmp_V[0];}
void IRAM_ATTR Bt_Enc() {digitalWrite(Beep, HIGH); Enc_SB(); Beep_ON = 1; Sec[3] = 0;}
void IRAM_ATTR Bt_Zar() {digitalWrite(Beep, HIGH); SB_Zar(); Beep_ON = 1; Sec[3] = 0;}
/*=======================================================================================================
                                              Инициализация
=======================================================================================================*/
void setup(void) 
{
  Serial.begin(115200);
  Wire.begin();

  Prefs.begin("Pref", false);
  Prefs.getBytes("Pref", Temp, sizeof(Temp));
  LOAD_R

  pinMode(Fan_PWM, OUTPUT);
  pinMode(PWM_OUT, OUTPUT);

  ledcSetup(Pwm_Channel, Pwm_Freq, Pwm_Bit);
  ledcAttachPin(PWM_OUT, Pwm_Channel);
  ledcAttachPin(Fan_PWM, Fan_Channel);
  
  pinMode(SB_Enc,  INPUT);
  pinMode(But_Zar, INPUT);
  attachInterrupt(SB_Enc,  Bt_Enc, RISING);
  attachInterrupt(But_Zar, Bt_Zar, RISING);

  u8g2.setI2CAddress(0x3F * 2);                       // адрес 0x3F!!!!!!!
  u8g2.setBusClock(400000);
  u8g2.begin();
  Serial.println(" << READY >> ");
  u8g2.setContrast(210);                              // Это констраст
  u8g2.setFlipMode(0);                                // Поворот изображения на 180 градусов
  Indic(0);
  delay(1000);

  pinMode(Beep, OUTPUT);
  digitalWrite(Beep, HIGH);
  delay(500);
  digitalWrite(Beep, LOW);
  
  encoder.attachHalfQuad (DT, CLK);
  encoder.setCount (0);
/*-----------------------------------------------------------------------------------------------------*/   
  PID_CUR.SetTunings(Kp[0], Ki[0], Kd[0]);
  PID_CUR.SetMode(AUTOMATIC);
  PID_CUR.SetOutputLimits(0, Var_L);
  PID_CUR.Compute();
  PID_ZAR.SetTunings(Kp[1], Ki[1], Kd[1]);
  PID_ZAR.SetMode(AUTOMATIC);
  PID_ZAR.SetOutputLimits(-Var_L, Var_L);
  PID_ZAR.Compute();
/*-----------------------------------------------------------------------------------------------------*/
  ads.setGain(GAIN_TWOTHIRDS);                          // Установка коэффициента усиления 2/3  

// ВОЗМОЖНЫЕ ВАРИАНТЫ УСТАНОВКИ КУ:                           
// ads.setGain(GAIN_TWOTHIRDS); | 2/3х | +/-6.144V | 1bit = 0.1875mV    |   
// ads.setGain(GAIN_ONE);       | 1х   | +/-4.096V | 1bit = 0.125mV     |  
// ads.setGain(GAIN_TWO);       | 2х   | +/-2.048V | 1bit = 0.0625mV    |  
// ads.setGain(GAIN_FOUR);      | 4х   | +/-1.024V | 1bit = 0.03125mV   |  
// ads.setGain(GAIN_EIGHT);     | 8х   | +/-0.512V | 1bit = 0.015625mV  |  
// ads.setGain(GAIN_SIXTEEN);   | 16х  | +/-0.256V | 1bit = 0.0078125mV |  

  ads.begin();                                          // Инициализация модуля ADS1115
/*-----------------------------------------------------------------------------------------------------*/
  Timer0_Cfg = timerBegin(0, 80, true);
  timerAttachInterrupt(Timer0_Cfg, &Timer0_ISR, true);
  timerAlarmWrite(Timer0_Cfg, 10000, true);
  timerAlarmEnable(Timer0_Cfg);
  Timer2_Cfg = timerBegin(2, 80, true);
  timerAttachInterrupt(Timer2_Cfg, &Timer2_ISR, true);
  timerAlarmWrite(Timer2_Cfg, 1000, true);
  timerAlarmEnable(Timer2_Cfg);
}
/*=======================================================================================================
                                              Главный цикл
=======================================================================================================*/
void loop(void) 
{
  if (Write == 1) WRITE();
  if (Flag_Cels == 1) Read_DS18B20();
  Read_Volt();
  if (Zar == 1) Zar_End();
  if (Sec[7] == 0) {Indic(1); Sec[7] = Ind_C;}
/*-----------------------------------------------------------------------------------------------------*/
  Sp_Pwm = map(Cels_BP[0], 0, 80, -1000, Var_L - 1);
  if ((Sp_Pwm <= 399) || (Sp_Pwm == Var_L)) Sp_Pwm = 400; 
  ledcWrite(Fan_Channel, Sp_Pwm);
  delay(10);
//    Serial.print(" : Sec_2 = "); Serial.println(Sec[2]);
}
/*=======================================================================================================
                                         Чтение напряжения и тока
=======================================================================================================*/
void Read_Volt() 
{
  Value_V(0); Volt[1] = Volt[1] + Volt_Value[0];
  Value_V(3); Curr[1] = Curr[1] + Volt_Value[3] / 10;
/*-----------------------------------------------------------------------------------------------------*/
  byte del = 20;
  Del_V++;
  if (Del_V >= del + 1) 
  {
    Volt[0] =  Volt[1] / del;
    Curr[0] = (Curr[1] / del) / 0.1;
    Volt[1] = 0.0; Curr[1] = 0.0; Del_V = 0;
  }
}
/*-----------------------------------------------------------------------------------------------------*/
void Value_V(byte Num)
{
  Adc[Num] = ads.readADC_SingleEnded(Num);
  Volt_Value[Num] = float(Adc[Num]) * 0.1875 / 100;
  if (Volt_Value[0] <= 0.08) Volt_Value[0] = 0.0;
  if (Volt_Value[3] <= 0.01) Volt_Value[3] = 0.0;
}
/*=======================================================================================================
                                              Button       // || - или, && - и
=======================================================================================================*/
void SB_Zar()
{
  if (Sec[9] == 0)
  {    
    switch(Zar)
    {
      case 0:
              if (Select == 10) {Write = 1; Select = 0; break;}
              if (Select == 2) {Level = Tmp_V[Select]; Zar = 2;}
              else
              {
                if (Select >= 3) 
                {
                  Zar = 1; Level = Volt_Min;
                  Curr_End = Curr[2];
                }
                else {Level = Tmp_V[0]; Select = 0;}
              }
      break;
      case 1:
              Zar = 0; Level = Tmp_V[0]; Ind_C = 30;
              Dw = 0; Control_PWM(0);
      break;
      case 2:
              Zar = 0; Level = Tmp_V[0]; Select = 0; Ind_C = 30;
              Dw = 0; Control_PWM(0); 
      break;
    }
    Beep_ON = 2;
    Sec[9] = 30;
  }
}
/*=======================================================================================================
                                              
=======================================================================================================*/
void Zar_End()                                             // Отключение режима зарядки
{
  if ((Select >= 3) && (Select <= 7))                     // Режим Литий отключение зарядки
  {
    if (Volt[0] >= Volt_Lim[Select]) 
    {
      Zar = 0; Select = 0; Level = Tmp_V[0]; Beep_ON = 1;
      Beep_OFF = 5; Sec[8] = 0; Ind_C = 30;
    }
  }
/*-----------------------------------------------------------------------------------------------------*/
  if (Select == 8)
  {
    if (Flag_Plum[0] == 0) {if (Volt[0] >= Volt_Lim[Select] - 2) Curr[2] /= 2; Flag_Plum[0] = 1;}
    if (Volt[0] >= Volt_Lim[Select]) 
    {
      Zar = 0; Select = 0; Level = Tmp_V[0]; Beep_ON = 1;
      Flag_Plum[0] = 0; Ind_C = 30; Beep_OFF = 5; Sec[8] = 0;
    }
  }
/*-----------------------------------------------------------------------------------------------------*/
  if (Select == 9) 
  {
    if (Flag_Plum[1] == 0)
    {
      if (Volt[0] >= Volt_Lim[Select]) 
      {
        if (Curr[0] >= Curr[2] - 0.1) Curr[2] -= 0.1; 
        Flag_Plum[1] = 1; Sec[3] = 20; Beep_ON = 1; Ind_C = 30;
      }
    }
    if (Curr[0] <= Curr[2]) Flag_Plum[1] = 0;
    if (Curr[2] <= Curr_End / 2) 
    {
      Zar = 0; Select = 0; Curr_End = 0; Ind_C = 30;
      Level = Tmp_V[0]; Beep_ON = 1; Beep_OFF = 5; Sec[8] = 0; Flag_Plum[1] = 0;
    }
  }
}
/*=======================================================================================================
                                              Encoder       // || - или, && - и
=======================================================================================================*/
void Encode()
{
  long newPosition = encoder.getCount() / 2;
  if (oldPosition != newPosition)
  {
    if (oldPosition < newPosition)
    {
      if (Select == 0) {if (Level <= Var_L) Level += 20; if (Level >= Var_L) Level = Var_L;}
      if ((Select >= 1) && (Select <= 7)) {if (Curr[2] >= 0.1) Curr[2] -= 0.1; Curr[3] = Curr[2]; Sec[5] = 0;}
      if (Select >= 8) {if (Curr[2] >= 0.1) Curr[2] -= 0.1; Curr[4] = Curr[2];}
    }
    if (oldPosition > newPosition)
    {
      if (Select == 0) {if (Level >= 2) Level -= 20; if (Level <= 1) Level = 1;}
      if ((Select >= 1) && (Select <= 7)) {if (Curr[2] <= 10.0) Curr[2] += 0.1; Curr[3] = Curr[2]; Sec[5] = 0;}
      if (Select >= 8) {if (Curr[2] <= 10.0) Curr[2] += 0.1; Curr[4] = Curr[2];}
    }
    oldPosition = newPosition;
    Serial.print(" : newPosition = "); Serial.println(newPosition);
  } 
}
/*=======================================================================================================
                                              Encoder
=======================================================================================================*/
void Enc_SB()
{
  if (Sec[9] == 0)
  {
    switch (Select)
    {
      case  0: Select = 1;  Curr[2] = Curr[3]; Sec[5] = 0; break;
      case  1: Select = 2;  Curr[2] = Curr[5]; break;
      case  2: Select = 3;  Curr[2] = Curr[3]; Volt_M(2); Control_PWM(0); break;
      case  3: Select = 4;  Curr[2] = Curr[3]; Volt_M(3); Control_PWM(0); break;
      case  4: Select = 5;  Curr[2] = Curr[3]; Volt_M(4); Control_PWM(0); break;
      case  5: Select = 6;  Curr[2] = Curr[3]; Volt_M(5); Control_PWM(0); break;
      case  6: Select = 7;  Curr[2] = Curr[3]; Volt_M(6); Control_PWM(0); break;
      case  7: Select = 8;  Curr[2] = Curr[4]; Volt_M(3); Control_PWM(0); break;
      case  8: Select = 9;  Curr[2] = Curr[3]; Volt_M(3); Control_PWM(0); break;
      case  9: Select = 10; Curr[2] = Curr[3]; break;
      case 10: Select = 0;  Curr[2] = Curr[3]; Level = Tmp_V[0]; Zar = 0; Zero_Time break;
    }
    Serial.print(" : Select = "); Serial.println(Select);
    Sec[9] = 30;
  }
}
/*=======================================================================================================

=======================================================================================================*/
void Control_PWM(uint8_t Sel)
{
  Input[Sel]    = Curr[2];
  Setpoint[Sel] = Curr[0];
  switch (Sel)
  {
    case 0: PID_CUR.Compute(); break;
    case 1: PID_ZAR.Compute(); break;
  }
  Duration = Level + Output[Sel];
  if (Duration >= Var_L) Duration = Var_L;
  if (Duration <= 1)     Duration = 1;
  ledcWrite(Pwm_Channel, Duration);
}
/*=======================================================================================================
                                              
=======================================================================================================*/
void WRITE()
{
  LOAD_W Prefs.putBytes("Pref", Temp, sizeof(Temp));
  Prefs.getBytes("Pref", Temp, sizeof(Temp)); LOAD_R Write = 0;
}
/*=======================================================================================================
                                              Чтение температуры
=======================================================================================================*/
void Read_DS18B20()
{
  if (FlagSensReady_BP == true)
  {
    FlagSensReady_BP = false;                                   // данные готовы

    if (OneWire::crc8(bufData_BP, 8) == bufData_BP[8])          // проверка CRC данные правильные
    {
      Cels_BP[0] = ((float)((int)bufData_BP[0] | (((int)bufData_BP[1]) << 8)) * 0.0625 + 0.03125);
    }
  }
  Flag_Cels = 0;
}
/*=====================================================================================================*/
